var __getOwnPropNames = Object.getOwnPropertyNames;
var __esm = (fn, res) => function __init() {
  return fn && (res = (0, fn[__getOwnPropNames(fn)[0]])(fn = 0)), res;
};
var __commonJS = (cb, mod) => function __require() {
  return mod || (0, cb[__getOwnPropNames(cb)[0]])((mod = { exports: {} }).exports, mod), mod.exports;
};

// src/seeder.ts
import BbPromise from "bluebird";
import _ from "lodash";
import path from "path";
import fs from "fs";
async function writeSeedBatch(dynamodbWriteFunction, tableName, seeds) {
  const params = {
    RequestItems: {
      [tableName]: seeds.map((seed) => ({
        PutRequest: {
          Item: seed
        }
      }))
    }
  };
  return new Promise((resolve, reject) => {
    let interval = 0;
    function execute(interval2) {
      setTimeout(() => dynamodbWriteFunction(params, (err) => {
        if (err) {
          if (err.code === "ResourceNotFoundException" && interval2 <= 5e3) {
            execute(interval2 + 1e3);
          } else {
            reject(err);
          }
        } else {
          resolve();
        }
      }), interval2);
    }
    execute(interval);
  });
}
function writeSeeds(dynamodbWriteFunction, tableName, seeds) {
  if (!dynamodbWriteFunction) {
    throw new Error("dynamodbWriteFunction argument must be provided");
  }
  if (!tableName) {
    throw new Error("table name argument must be provided");
  }
  if (!seeds) {
    throw new Error("seeds argument must be provided");
  }
  if (seeds.length > 0) {
    const seedChunks = _.chunk(seeds, MAX_MIGRATION_CHUNK);
    return BbPromise.map(
      seedChunks,
      (chunk) => writeSeedBatch(dynamodbWriteFunction, tableName, chunk),
      { concurrency: MIGRATION_SEED_CONCURRENCY }
    ).then(() => console.log("Seed running complete for table: " + tableName));
  }
  return Promise.resolve();
}
function fileExists(fileName) {
  return new Promise((resolve) => {
    fs.access(fileName, (exists) => resolve(exists));
  });
}
function unmarshalBuffer(json) {
  _.forEach(json, function(value, key) {
    if (value !== null && value.type === "Buffer") {
      json[key] = Buffer.from(value.data);
    }
  });
  return json;
}
function getSeedsAtLocation(location) {
  const result = JSON.parse(fs.readFileSync(location, "utf8"));
  if (Array.isArray(result)) {
    return _.forEach(result, unmarshalBuffer);
  } else {
    return [unmarshalBuffer(result)];
  }
}
function locateSeeds(sources = [], cwd = process.cwd()) {
  const locations = sources.map((source) => path.join(cwd, source));
  return BbPromise.map(locations, (location) => {
    return fileExists(location).then((exists) => {
      if (!exists) {
        throw new Error("source file " + location + " does not exist");
      }
      return getSeedsAtLocation(location);
    });
  }).then(_.flatten);
}
var MAX_MIGRATION_CHUNK, MIGRATION_SEED_CONCURRENCY;
var init_seeder = __esm({
  "src/seeder.ts"() {
    "use strict";
    MAX_MIGRATION_CHUNK = 25;
    MIGRATION_SEED_CONCURRENCY = 5;
  }
});

// src/index.ts
import _2 from "lodash";
import BbPromise2 from "bluebird";
import AWS from "aws-sdk";
import dynamodbLocal from "dynamodb-localhost";
import path2 from "path";
var require_src = __commonJS({
  "src/index.ts"(exports, module) {
    init_seeder();
    var ServerlessDynamodbLocal = class {
      constructor(serverless, options, logging) {
        this.serverless = serverless;
        this.logging = logging;
        this.serverless = serverless;
        this.service = serverless.service;
        this.config = this.service.custom?.["dynamodb"] ?? {};
        const localPath = serverless.config && path2.join(serverless.config.servicePath, ".dynamodb");
        this.options = {
          ...options,
          localPath: options.localPath ?? localPath
        };
      }
      config;
      service;
      options;
      provider = "aws";
      hooks = {
        "dynamodb:migrate:migrateHandler": this.migrateHandler.bind(this),
        "dynamodb:seed:seedHandler": this.seedHandler.bind(this),
        "dynamodb:remove:removeHandler": this.removeHandler.bind(this),
        "dynamodb:install:installHandler": this.installHandler.bind(this),
        "dynamodb:start:startHandler": this.startHandler.bind(this),
        "before:offline:start:init": this.startHandler.bind(this),
        "before:offline:start:end": this.endHandler.bind(this)
      };
      commands = {
        dynamodb: {
          commands: {
            migrate: {
              lifecycleEvents: ["migrateHandler"],
              usage: "Creates local DynamoDB tables from the current Serverless configuration",
              options: {
                online: {
                  shortcut: "o",
                  usage: "Uses online tables (default is offline)",
                  type: "boolean"
                },
                convertEmptyValues: {
                  shortcut: "e",
                  usage: "Set to true if you would like the document client to convert empty values (0-length strings, binary buffers, and sets) to be converted to NULL types when persisting to DynamoDB.",
                  type: "boolean"
                }
              }
            },
            start: {
              lifecycleEvents: ["startHandler"],
              usage: "Starts local DynamoDB",
              options: {
                port: {
                  shortcut: "p",
                  usage: "The port number that DynamoDB will use to communicate with your application. If you do not specify this option, the default port is 8000",
                  type: "string"
                },
                cors: {
                  shortcut: "c",
                  usage: 'Enable CORS support (cross-origin resource sharing) for JavaScript. You must provide a comma-separated "allow" list of specific domains. The default setting for -cors is an asterisk (*), which allows public access.',
                  type: "string"
                },
                inMemory: {
                  shortcut: "i",
                  usage: "DynamoDB; will run in memory, instead of using a database file. When you stop DynamoDB;, none of the data will be saved. Note that you cannot specify both -dbPath and -inMemory at once.",
                  type: "boolean"
                },
                dbPath: {
                  shortcut: "d",
                  usage: "The directory where DynamoDB will write its database file. If you do not specify this option, the file will be written to the current directory. Note that you cannot specify both -dbPath and -inMemory at once. For the path, current working directory is <projectroot>/node_modules/serverless-dynamodb-local/dynamob. For example to create <projectroot>/node_modules/serverless-dynamodb-local/dynamob/<mypath> you should specify -d <mypath>/ or --dbPath <mypath>/ with a forwardslash at the end.",
                  type: "string"
                },
                sharedDb: {
                  shortcut: "h",
                  usage: "DynamoDB will use a single database file, instead of using separate files for each credential and region. If you specify -sharedDb, all DynamoDB clients will interact with the same set of tables regardless of their region and credential configuration.",
                  type: "boolean"
                },
                delayTransientStatuses: {
                  shortcut: "t",
                  usage: "Causes DynamoDB to introduce delays for certain operations. DynamoDB can perform some tasks almost instantaneously, such as create/update/delete operations on tables and indexes; however, the actual DynamoDB service requires more time for these tasks. Setting this parameter helps DynamoDB simulate the behavior of the Amazon DynamoDB web service more closely. (Currently, this parameter introduces delays only for global secondary indexes that are in either CREATING or DELETING status.",
                  type: "boolean"
                },
                optimizeDbBeforeStartup: {
                  shortcut: "o",
                  usage: "Optimizes the underlying database tables before starting up DynamoDB on your computer. You must also specify -dbPath when you use this parameter.",
                  type: "boolean"
                },
                migrate: {
                  shortcut: "m",
                  usage: "After starting dynamodb local, create DynamoDB tables from the current serverless configuration.",
                  type: "boolean"
                },
                seed: {
                  shortcut: "s",
                  usage: "After starting and migrating dynamodb local, injects seed data into your tables. The --seed option determines which data categories to onload.",
                  type: "string"
                },
                heapInitial: {
                  usage: 'The initial heap size. Specify megabytes, gigabytes or terabytes using m, b, t. E.g., "2m"',
                  type: "string"
                },
                heapMax: {
                  usage: 'The maximum heap size. Specify megabytes, gigabytes or terabytes using m, b, t. E.g., "2m"',
                  type: "string"
                },
                convertEmptyValues: {
                  shortcut: "e",
                  usage: "Set to true if you would like the document client to convert empty values (0-length strings, binary buffers, and sets) to be converted to NULL types when persisting to DynamoDB.",
                  type: "boolean"
                },
                docker: {
                  usage: "Run DynamoDB inside docker container instead of as a local Java program",
                  type: "boolean"
                },
                host: {
                  usage: "The host name to use when connecting to DynamoDB. Defaults to localhost.",
                  type: "string"
                },
                online: {
                  shortcut: "o",
                  usage: "Uses online tables for migration and seed (default is offline)",
                  type: "boolean"
                },
                noStart: {
                  shortcut: "n",
                  usage: "Do not start DynamoDB local (in case it is already running)",
                  type: "boolean"
                }
              }
            },
            remove: {
              lifecycleEvents: ["removeHandler"],
              usage: "Removes local DynamoDB"
            },
            install: {
              usage: "Installs local DynamoDB",
              lifecycleEvents: ["installHandler"],
              options: {
                localPath: {
                  shortcut: "x",
                  usage: "Local dynamodb install path",
                  type: "string"
                }
              }
            }
          }
        }
      };
      get port() {
        return Number.parseInt(this.config.start.port ?? "8000");
      }
      get host() {
        return this.config.start.host ?? "localhost";
      }
      /**
       * Get the stage
       *
       * @return {String} the current stage
       */
      get stage() {
        return this.options && this.options.stage || this.service.provider && this.service.provider.stage;
      }
      /**
       * To check if the handler needs to be executed based on stage
       *
       * @return {Boolean} if the handler can run for the provided stage
       */
      shouldExecute() {
        if (this.config.stages?.includes(this.stage)) {
          return true;
        }
        return false;
      }
      dynamodbOptions(options) {
        let dynamoOptions = {};
        if (options?.online) {
          this.serverless.cli.log("Connecting to online tables...");
          if (!options.region) {
            throw new Error("please specify the region");
          }
          dynamoOptions = {
            region: options.region,
            convertEmptyValues: options?.convertEmptyValues ?? false
          };
        } else {
          dynamoOptions = {
            endpoint: `http://${this.host}:${this.port}`,
            region: options?.region ?? process.env["AWS_REGION"] ?? "localhost",
            accessKeyId: process.env["AWS_ACCESS_KEY_ID"] ?? "MOCK_ACCESS_KEY_ID",
            secretAccessKey: process.env["AWS_SECRET_ACCESS_KEY"] ?? "MOCK_SECRET_ACCESS_KEY",
            convertEmptyValues: options?.convertEmptyValues ?? false
          };
        }
        return {
          raw: new AWS.DynamoDB(dynamoOptions),
          doc: new AWS.DynamoDB.DocumentClient(dynamoOptions)
        };
      }
      migrateHandler() {
        if (this.shouldExecute()) {
          const dynamodb = this.dynamodbOptions();
          return Promise.all(this.tables.map((table) => this.createTable(dynamodb, table)));
        } else {
          this.serverless.cli.log("Skipping migration: DynamoDB Local is not available for stage: " + this.stage);
          return Promise.resolve();
        }
      }
      seedHandler() {
        if (this.shouldExecute()) {
          const options = this.options;
          const dynamodb = this.dynamodbOptions(options);
          return BbPromise2.each(this.seedSources, (source) => {
            if (!source?.table) {
              throw new Error('seeding source "table" property not defined');
            }
            const seedPromise = locateSeeds(source.sources || []).then((seeds) => writeSeeds(dynamodb.doc.batchWrite.bind(dynamodb.doc), source.table, seeds));
            const rawSeedPromise = locateSeeds(source?.rawsources || []).then((seeds) => writeSeeds(dynamodb.raw.batchWriteItem.bind(dynamodb.raw), source.table, seeds));
            return BbPromise2.all([seedPromise, rawSeedPromise]);
          });
        } else {
          this.serverless.cli.log("Skipping seeding: DynamoDB Local is not available for stage: " + this.stage);
          return Promise.resolve();
        }
      }
      removeHandler() {
        return new BbPromise2((resolve) => dynamodbLocal.remove(resolve));
      }
      installHandler() {
        const options = this.options;
        return new BbPromise2((resolve) => dynamodbLocal.install(resolve, options.localPath));
      }
      startHandler() {
        if (this.shouldExecute()) {
          const options = {
            ...this.config.start,
            ...this.options,
            sharedDb: this.options.sharedDb ?? true,
            install_path: this.options.localPath
          };
          let dbPath = options.dbPath;
          if (dbPath) {
            options.dbPath = path2.isAbsolute(dbPath) ? dbPath : path2.join(this.serverless.config.servicePath, dbPath);
          }
          if (!this.config.start.noStart) {
            dynamodbLocal.start({
              port: Number.parseInt(options.port ?? "8000"),
              cors: options.cors,
              inMemory: options.inMemory,
              dbPath: options.dbPath,
              docker: options.docker,
              sharedDb: options.sharedDb,
              delayTransientStatuses: options.delayTransientStatuses,
              optimizeDbBeforeStartup: options.optimizeDbBeforeStartup,
              heapInitial: options.heapInitial,
              heapMax: options.heapMax
            });
          }
          return BbPromise2.resolve().then(async () => options.migrate && await this.migrateHandler()).then(async () => options.seed && await this.seedHandler());
        } else {
          this.serverless.cli.log("Skipping start: DynamoDB Local is not available for stage: " + this.stage);
          return Promise.resolve();
        }
      }
      endHandler() {
        if (this.shouldExecute() && !this.config.start.noStart) {
          this.serverless.cli.log("DynamoDB - stopping local database");
          dynamodbLocal.stop(this.port);
        } else {
          this.serverless.cli.log("Skipping end: DynamoDB Local is not available for stage: " + this.stage);
        }
      }
      getDefaultStack() {
        return _2.get(this.service, "resources");
      }
      getAdditionalStacks() {
        return _2.values(_2.get(this.service, "custom.additionalStacks", {}));
      }
      hasAdditionalStacksPlugin() {
        return this.service.plugins?.includes("serverless-plugin-additional-stacks") ?? false;
      }
      getTableDefinitionsFromStack(stack) {
        const resources = stack?.["Resources"] ?? [];
        return Object.keys(resources).filter((key) => resources[key]?.["Type"] === "AWS::DynamoDB::Table" && resources[key]?.["Properties"]).map((key) => resources[key]?.["Properties"]);
      }
      /**
       * Gets the table definitions
       */
      get tables() {
        let stacks = [];
        const defaultStack = this.getDefaultStack();
        if (defaultStack) {
          stacks.push(defaultStack);
        }
        if (this.hasAdditionalStacksPlugin()) {
          stacks = stacks.concat(this.getAdditionalStacks());
        }
        return stacks.map((stack) => this.getTableDefinitionsFromStack(stack)).reduce((tables, tablesInStack) => tables.concat(tablesInStack), []);
      }
      /**
       * Gets the seeding sources
       */
      get seedSources() {
        const seedConfig = this.config.seed ?? {};
        const seed = this.options.seed || this.config.start.seed || seedConfig;
        const categories = [];
        if (typeof seed === "string") {
          categories.push(...seed.split(","));
        } else if (seed) {
          categories.push(...Object.keys(seedConfig));
        } else {
          this.serverless.cli.log("DynamoDB - No seeding defined. Skipping data seeding.");
          return [];
        }
        const sourcesByCategory = categories.map((category) => seedConfig[category]?.sources);
        return _2.flatten(sourcesByCategory);
      }
      async createTable(dynamodb, cfnTable) {
        if (!cfnTable.TableName || !cfnTable.KeySchema) {
          throw new Error("Table name, key schema and attribute definitions are required");
        }
        const migration = {
          TableName: cfnTable.TableName,
          KeySchema: cfnTable.KeySchema,
          AttributeDefinitions: cfnTable.AttributeDefinitions ?? [],
          LocalSecondaryIndexes: cfnTable.LocalSecondaryIndexes,
          GlobalSecondaryIndexes: cfnTable.GlobalSecondaryIndexes,
          ProvisionedThroughput: cfnTable.ProvisionedThroughput,
          Tags: cfnTable.Tags
        };
        if (cfnTable.StreamSpecification && cfnTable.StreamSpecification.StreamViewType) {
          migration.StreamSpecification = {
            ...cfnTable.StreamSpecification,
            StreamEnabled: true
          };
        }
        if (cfnTable.SSESpecification) {
          migration.SSESpecification = {
            Enabled: cfnTable.SSESpecification.SSEEnabled
          };
        }
        if (!migration.BillingMode || migration.BillingMode === "PAY_PER_REQUEST") {
          delete migration.BillingMode;
          const defaultProvisioning = {
            ReadCapacityUnits: 5,
            WriteCapacityUnits: 5
          };
          migration.ProvisionedThroughput = defaultProvisioning;
          if (migration.GlobalSecondaryIndexes) {
            migration.GlobalSecondaryIndexes.forEach((gsi) => {
              gsi.ProvisionedThroughput = defaultProvisioning;
            });
          }
        }
        try {
          await dynamodb.raw.createTable(migration).promise();
          this.serverless.cli.log("DynamoDB - created table " + migration.TableName);
          return migration;
        } catch (err) {
          if (err instanceof Error && err.name === "ResourceInUseException") {
            this.serverless.cli.log(`DynamoDB - Warn - table ${migration.TableName} already exists`);
            return;
          } else {
            this.serverless.cli.log("DynamoDB - Error - ", String(err));
            throw err;
          }
        }
      }
    };
    module.exports = ServerlessDynamodbLocal;
  }
});
export default require_src();
